import argparse , math
import Qlearners.recipes.cartpole.cartpolebox2d as cartpole
import matplotlib.pyplot as plt
import numpy as np
import Qlearners.Qlearner as ql

parser = argparse.ArgumentParser()
parser.add_argument( "-v", action="store_true", default=False, help="not implemented")
parser.add_argument( "--M_H", nargs="+", type=int, default=[20,20])
parser.add_argument( "--scgI", type=int, default=20)
parser.add_argument( "--gamma", type=float, default=0.9)
parser.add_argument( "--rerunNum", type=int)
parser.add_argument( "--numReplays", type=int, default=5)
parser.add_argument( "--episode_len", type=int, default=1000)
parser.add_argument( "--episodes_max", type=int, default=200)
parser.add_argument( "--batch_size", type=int, default=1000)
parser.add_argument( "--graph",action="store_true", default=False)
parser.add_argument( "--saveEvalHist",action="store_true", default=False)
parser.add_argument( "--saveWeightsInterval", type=int, default=0, help="how often to save weights.  zero for don't save. Value of one will save every eval.")
parser.add_argument( "--evalLength" , type=int , default=2000)
parser.add_argument( "--saveDir")
parser.add_argument( "--savePrefix")
args = parser.parse_args()

if args.saveEvalHist:
    saveEvalFile = tempfile.NamedTemporaryFile(mode="w",delete=False,
                                               dir=args.saveDir,
                                               prefix=args.savePrefix,
                                               suffix=".outfile")

actions = ( [-1.0] , [0.0] , [1.0] )

#
# prepare plotting
#

plt.ion(); firstGraph=True;

#
# setup some constant parameters
#

eval_startPos = [0]                  # start eval in middle only
eval_startAngles = [-math.pi,0.0]    # start eval on up and down positions respectively
eval_r_hist = [[] for i in range(len(eval_startPos)*len(eval_startAngles))]
actionDuration = 2

#
# method to scale the function appoximator inputs
# 
def scaleState(tmpI):
    tmpI[:,0] = tmpI[:,0] / scaleState.positionRange[1]     
    tmpI[:,1] = tmpI[:,1] / scaleState.velocityRange[1]     
    tmpI[:,2] = tmpI[:,2] / scaleState.angleRange[1]
    tmpI[:,3] = tmpI[:,3] / scaleState.angleVelocityRange[1]
    return(tmpI)
scaleState.positionRange = (-2.2,2.2)
scaleState.velocityRange = (-6.,6.)
scaleState.angleVelocityRange = (-14.,14.)
scaleState.angleRange = (-math.pi,math.pi)

#
# method (and helper method) to generate 
# 
def reinforcement(angle):
    angle = abs(angle)
    if angle > math.pi * 0.75:
        return -1
    elif angle < math.pi * 0.25:
        return 1
    else:
        return 0

# ####################
# initialize cartpole environment
#
def get_domain():
    return get_domain.cpdomain
get_domain.cpdomain = cartpole.CartPole()

#
# ####################

# ####################
# create functions needed by Qlearner
# initialize Q-learning agent
#

def advance_environment_f( selected_action_idx , domain ):

    # move simulation forward, get new state, scale state, return scaled state
    for ai in range(actionDuration):
        domain.act( selected_action[selected_action_idx][0] )
    (x,xdot,a,adot) = domain.sense()
    # compute reward
    reward = reinforcement( a )
    # scale state
    state_vec = np.array( [ [x,xdot,a,adot] ] )
    scaleState( state_vec )
    # return scaled state and reward and indicate not done (sim always not done)
    return ( state_vec , reward , False )

def reset_domain_f():

    # doesn't actually reset: just picks up where previous episode left off
    # if want to start from same position, need to create a new instance of simulation
    
    domain = get_domain()
    (x,xdot,a,adot) = domain.sense()
    state_vec = np.array( [ [x,xdot,a,adot] ] )
    scaleState( state_vec )
    return state_vec


def reset_domain_eval_f( domain ):

    (x,xdot,a,adot) = domain.sense()
    state_vec = np.array( [ [x,xdot,a,adot] ] )
    scaleState( state_vec )
    return state_vec


agent = ql.Qlearner( 4 , actions , args.M_H , gamma=args.gamma )

# 
# ####################

# ####################
# loop over episodes
# 

episode_i = 0
while episode_i < args.episodes_max:

    episode = agent.generate_episode( args.episode_len ,
                                      step_f = lambda sa: advance_environment_f(sa , get_domain() ) ,
                                      init_f = reset_domain_f ,
                                      epsilon = 0.1 )
    agent.add_to_memory( episode )
    agent.learn( num_updates=args.numReplays , batch_size=args.batch_size )


    eval_domain = cartpole.CartPole()
    eval_episode = agent.generate_episode( args.evalLength ,
                                           step_f = lambda sa: advance_environment_f( sa , eval_domain ) ,
                                           init_f = lambda: reset_domain_eval_f( eval_domain ) ,
                                           epsilon = None )
    
    eval_sum_r = np.sum( [ ee["r"] for ee in eval_episode ] )

    print( "episode" , episode_i , "eval reward=" , eval_sum_r )

    episode_i += 1
    
# 
# ####################



"""
ANN class for Q-function approximation
"""

import functools
import torch
import torch.nn as nn
from torch import Tensor, LongTensor, optim
import Qlearners.scg as scg

class Net(nn.Module):
    """
    A totally over-built ANN class.
    """
    def __init__( self, M_I , M_H ,
                  M_H_Q=[] , M_F_O=0 ,
                  scgI=20 ,
                  noQonShared=False ):
        super(Net, self).__init__()
        self.M_I_Q = M_I ; self.M_H = M_H ; self.M_H_Q = M_H_Q ;
        self.Qlayers = []; self.Flayers = [];
        self._scgI = scgI
        self.noQonShared = noQonShared
        # create shared hidden layers
        self.h0 = nn.Linear(self.M_I_Q,self.M_H[0])
        if not self.noQonShared:
            self.Qlayers += ["h0"]
        self.Flayers += ["h0"]
        for li in range(1,len(self.M_H)):
            attrName = "h"+str(li)
            setattr(self, attrName,
                    nn.Linear(self.M_H[li-1],self.M_H[li]))
            if not self.noQonShared:
                self.Qlayers += [attrName]
            self.Flayers += [attrName]
        # create Q-learning hidden layers
        if self.M_H_Q:
            self.qh0 = nn.Linear(self.M_H[-1],self.M_H_Q[0])
            self.Qlayers += ["qh0"]
            for li in range(1,len(self.M_H_Q)):
                attrName = "qh"+str(li)
                setattr(self, attrName,
                        nn.Linear(self.M_H_Q[li-1],self.M_H_Q[li]))
                self.Qlayers += [attrName]
        # create Q-learning output layer
        if self.M_H_Q:
            self.qout = nn.Linear(self.M_H_Q[-1],1)
        else:
            self.qout = nn.Linear(self.M_H[-1],1)
        self.Qlayers += ["qout"]
        # create forward model output layer
        self.fout = nn.Linear(self.M_H[-1],M_F_O)
        self.Flayers += ["fout"]
        
    def compute_Q( self , x ):
        x = torch.tensor( x , dtype=torch.float , requires_grad=False )
        Q = self.forward_Q( x , grad_p=False )
        return Q.numpy()

    def forward_Q( self , x , grad_p=True):
        with torch.set_grad_enabled( grad_p ):
            for li in range(len(self.M_H)):
                layer = getattr(self,"h"+str(li))
                x = torch.tanh(layer(x))
            for li in range(len(self.M_H_Q)):
                layer = getattr(self,"qh"+str(li))
                x = torch.tanh(layer(x))
            x = self.qout(x)
        return x

    # def forward_F(self, x):
    #     for li in range(len(self.M_H)):
    #         layer = getattr(self,"h"+str(li))
    #         x = torch.tanh(layer(x))
    #     x = torch.tanh(self.fout(x))
    #     return x

    def _Q_parameters(self):
        return iter(functools.reduce(lambda x,y: x+y, [list(mod.parameters()) for name,mod in self.named_children() if name in self.Qlayers]))

    def turnOff_H_updates(self):
        if self.M_H:
            for li in range(0,len(self.M_H)):            
                layer = getattr(self,"h"+str(li))
                for param in layer.parameters():
                    param.requires_grad = False


    def update( self , state_action_in , targets_in , alg="scg" ):

        # TODO: check inputs

        N_in , M_in = state_action_in.shape
        N_out , M_out = targets_in.shape
        
        if alg=="scg":
            self._update_scg( state_action_in , targets_in , self._scgI )
        else:
            print( "Unknown ANN update algorithm." )
            exit

    def _update_scg( self , state_action_in , targets_in , num_scg_iters):

        #
        # define closure needed by scg
        #
        def closure(grad=False):
            tmpOut = self.forward_Q( torch.tensor( state_action_in , dtype=torch.float , requires_grad=grad ) , grad_p=grad )
            delta = tmpOut.clone()
            delta.sub_( torch.tensor( targets_in , dtype=torch.float ) )
            err = delta.pow(2.)
            err = 0.5 * torch.mean(err)
            if grad:
                delta.div_(tmpOut.size()[0]-1)
                self.zero_grad()
                tmpOut.backward(gradient=delta)
            return err
        #
        # run SCG for the specified number of iterations
        # 
        optimizer = scg.SCG(self._Q_parameters())
        for scgi in range(num_scg_iters):
            step_p = optimizer.step(closure=closure)
